# =============================================================================
# Minet Youtube Comments CLI Action
# =============================================================================
#
# From a video id, action getting all commments' data using Google's APIs.
#
import time
import csv
from tqdm import tqdm
from minet.cli.youtube.utils import seconds_to_midnight_pacific_time
from minet.cli.utils import die, open_output_file
from minet.utils import create_pool, request_json

URL_TEMPLATE = 'https://www.googleapis.com/youtube/v3/commentThreads?videoId=%(id)s&key=%(key)s&part=snippet,replies&maxResults=100'

CSV_HEADERS = [
    'video_id',
    'comment_id',
    'author_name',
    'author_channel_url',
    'author_channel_id',
    'text',
    'like_count',
    'published_at',
    'updated_at',
    'total_reply',
    'reply_to'
]


def get_data(data_json):
    data = []
    data_replies = []

    next_page = data_json.get('nextPageToken', None)
    all_items = data_json.get('items', None)
    is_reply = False

    if not all_items:
        all_items = data_json.get('comments', None)
        is_reply = True

    for item in all_items:

        comment_id = item.get('id', None)
        snippet = item['snippet']

        if is_reply:
            comment_data = snippet
        else:
            replies = item.get('replies', None)

            if replies:
                _, data_replies = get_data(replies)
                data.extend(data_replies)

            top_comment = snippet.get('topLevelComment', None)
            comment_data = top_comment.get('snippet', None)

        total_reply = snippet.get('totalReplyCount', None)
        author_name = comment_data['authorDisplayName']
        author_channel_url = comment_data['authorChannelUrl']
        author_channel_id = comment_data['authorChannelId']['value']
        video_id = comment_data['videoId']
        text = comment_data['textOriginal']
        like_count = comment_data['likeCount']
        published_at = comment_data['publishedAt']
        updated_at = comment_data['updatedAt']
        reply_to = comment_data.get('parentId', None)

        data.append([
            video_id,
            comment_id,
            author_name,
            author_channel_url,
            author_channel_id,
            text,
            like_count,
            published_at,
            updated_at,
            total_reply,
            reply_to
        ])

    return next_page, data


def comments_action(namespace, output_file):

    output_file = open_output_file(namespace.output)

    writer = csv.writer(output_file)
    writer.writerow(CSV_HEADERS)

    loading_bar = tqdm(
        desc='Retrieving',
        dynamic_ncols=True,
        unit=' comments',
    )

    http = create_pool()

    url = URL_TEMPLATE % {'id': namespace.id, 'key': namespace.key}
    next_page = True
    all_data = []

    while next_page:

        if next_page is True:
            err, response, result = request_json(http, url)
        else:
            url_next = url + '&pageToken=' + next_page
            err, response, result = request_json(http, url_next)

        if err:
            die(err)
        elif response.status == 403:
            time.sleep(seconds_to_midnight_pacific_time())
            continue
        elif response.status >= 400:
            die(response.status)

        next_page, data = get_data(result)

        for comment in data:
            loading_bar.update()
            writer.writerow(comment)



# =============================================================================
# Minet Twitter Users CLI Action
# =============================================================================
#
# Logic of the `tw users` action.

import casanova
from minet.twitter.utils import TwitterWrapper, clean_user_entities, get_timestamp
from minet.utils import chunks_iter
from tqdm import tqdm
import datetime

REPORT_HEADERS = [
    'id_str',
    'screen_name',
    'name',
    'location',
    'created_at_utc',
    'created_at_iso_local',
    'description',
    'url',
    'protected',
    'verified',
    'followers_count',
    'friends_count',
    'favourites_count',
    'listed_count',
    'statuses_count',
    'default_profile',
    'default_profile_image',
    'profile_image_url_https',
    'profile_banner_url'
]


def get_data(result, key):

    data_indexed = {}
    locale = datetime.datetime.now(datetime.timezone(datetime.timedelta(0))).astimezone().tzinfo

    for element in result:
        clean_user_entities(element)

        user_index = element.get(key)

        def getter(element, k):
            if k != 'created_at_iso_local':
                return element.get(k)
            else:
                return get_timestamp(element.get('created_at'), locale)

        data = [getter(element, key) for key in REPORT_HEADERS]

        data_indexed[user_index] = data

    return data_indexed


def twitter_users_action(namespace, output_file):

    TWITTER = {
        'access_token': namespace.access_token,
        'access_token_secret': namespace.access_token_secret,
        'api_key': namespace.api_key,
        'api_secret_key': namespace.api_secret_key
    }

    wrapper = TwitterWrapper(TWITTER)

    enricher = casanova.enricher(
        namespace.file,
        output_file,
        keep=namespace.select,
        add=REPORT_HEADERS
    )

    loading_bar = tqdm(
        desc='Retrieving ids',
        dynamic_ncols=True,
        total=namespace.total,
        unit=' line'
    )

    for chunk in chunks_iter(enricher.cells(namespace.column, with_rows=True), 100):
        users = ','.join(row[1] for row in chunk)

        if namespace.id:
            wrapper_args = {'user_id': users}
            key = 'id_str'
        else:
            wrapper_args = {'screen_name': users}
            key = 'screen_name'

        result = wrapper.call('users.lookup', wrapper_args)

        if result is not None:
            data = get_data(result, key)

            for row, user in chunk:
                if user in data.keys():
                    enricher.writerow(row, data[user])
                else:
                    enricher.writerow(row)

        loading_bar.update(len(chunk))

    loading_bar.close()



# =============================================================================
# Minet Mediacloud Utils
# =============================================================================
#
# Miscellaneous utility function used throughout the mediacloud package.
#


def get_next_link_id(data):

    if 'link_ids' not in data:
        return None

    pagination = data['link_ids']

    if not pagination.get('next'):
        return None

    return pagination['next']


def explode_tags(data, join_char=None):
    tags = []
    tag_sets = []
    tags_ids = []
    tag_sets_ids = []

    for tag in data:
        tags.append(tag['tag'])
        tag_sets.append(tag['tag_set'])
        tags_ids.append(tag['tags_id'])
        tag_sets_ids.append(tag['tag_sets_id'])

    if join_char is not None:
        tags = join_char.join(tags)
        tag_sets = join_char.join(tag_sets)
        tags_ids = join_char.join(str(i) for i in tags_ids)
        tag_sets_ids = join_char.join(str(i) for i in tag_sets_ids)

    return tags, tag_sets, tags_ids, tag_sets_ids


def get_last_processed_stories_id(data):
    if not data:
        return None

    return data[-1]['processed_stories_id']



# =============================================================================
# Minet CrowdTangle Links Summary CLI Action
# =============================================================================
#
# Logic of the `ct summary` action.
#
import csv
import casanova
from io import StringIO
from tqdm import tqdm
from ural import is_url

from minet.cli.utils import die
from minet.crowdtangle.constants import (
    CROWDTANGLE_SUMMARY_CSV_HEADERS,
    CROWDTANGLE_POST_CSV_HEADERS_WITH_LINK
)
from minet.crowdtangle.exceptions import (
    CrowdTangleInvalidTokenError
)
from minet.crowdtangle import CrowdTangleClient


def crowdtangle_summary_action(namespace, output_file):
    if not namespace.start_date:
        die('Missing --start-date!')

    if is_url(namespace.column):
        namespace.file = StringIO('url\n%s' % namespace.column)
        namespace.column = 'url'

    enricher = casanova.enricher(
        namespace.file,
        output_file,
        keep=namespace.select.split(',') if namespace.select else None,
        add=CROWDTANGLE_SUMMARY_CSV_HEADERS
    )

    posts_writer = None

    if namespace.posts is not None:
        posts_writer = csv.writer(namespace.posts)
        posts_writer.writerow(CROWDTANGLE_POST_CSV_HEADERS_WITH_LINK)

    loading_bar = tqdm(
        desc='Collecting data',
        dynamic_ncols=True,
        total=namespace.total,
        unit=' urls'
    )

    client = CrowdTangleClient(namespace.token, rate_limit=namespace.rate_limit)

    for row, url in enricher.cells(namespace.column, with_rows=True):
        url = url.strip()

        try:
            stats = client.summary(
                url,
                start_date=namespace.start_date,
                with_top_posts=namespace.posts is not None,
                sort_by=namespace.sort_by,
                format='csv_row'
            )

        except CrowdTangleInvalidTokenError:
            die([
                'Your API token is invalid.',
                'Check that you indicated a valid one using the `--token` argument.'
            ])

        except Exception as err:
            raise err

        if namespace.posts is not None:
            stats, posts = stats

            if posts is not None:
                for post in posts:
                    posts_writer.writerow([url] + post)

        enricher.writerow(row, stats)

        loading_bar.update()



# =============================================================================
# Minet Facebook Post Stats CLI Action
# =============================================================================
#
# Logic of the `fb post-stats` action.
#
import re
import json5
import casanova
from tqdm import tqdm
from bs4 import BeautifulSoup
from datetime import datetime
from collections import OrderedDict
from ural.facebook import is_facebook_url

from minet.utils import create_pool, request, nested_get, sleep_with_entropy
from minet.cli.utils import open_output_file, print_err, die
from minet.cli.facebook.constants import FACEBOOK_WEB_DEFAULT_THROTTLE

META_EXTRACTOR_TEMPLATE = rb'\(function\(\)\{bigPipe\.onPageletArrive\((\{.+share_fbid:"%s".+\})\);\}\),"onPageletArrive'
HTML_EXTRACTOR_TEMPLATE = rb'<code[^>]*><!--(.+%s.+)--></code>'
CURRENT_AVAILABILITY_DISCLAIMER = b'The link you followed may have expired, or the page may only be visible to an audience'
AVAILABILITY_DISCLAIMER = b'The link you followed may be broken, or the page may have been removed'
LOGIN_DISCLAIMER = b'You must log in to continue'
CAPTCHA = b'id="captcha"'

REPORT_HEADERS = [
    'error',
    'canonical',
    'account_name',
    'timestamp',
    'time',
    'link',
    'aria_label',
    'text',
    'share_count',
    'comment_count',
    'reaction_count',
    'video_view_count'
]

REACTION_KEYS = OrderedDict({
    1: 'like',
    2: 'love',
    3: 'wow',
    4: 'haha',
    7: 'sad',
    8: 'angry',
    11: 'thankful',
    12: 'pride'
})

for emotion_name in REACTION_KEYS.values():
    REPORT_HEADERS.append('%s_count' % emotion_name)

ERROR_PADDING = [''] * (len(REPORT_HEADERS) - 1)


# TODO: integrate to ural
def is_facebook_post_url(url):
    return '/posts/' in url or '/permalink/' in url


def get_count(item):
    if item is None:
        return 0

    value = item.get('count')

    if not value:
        value = item.get('total_count')

    return value or 0


def collect_top_reactions(data):
    edges = nested_get(['top_reactions', 'edges'], data)

    if edges is None:
        return

    index = {}

    for edge in edges:
        emotion = REACTION_KEYS.get(edge['node']['key'])

        if emotion is None:
            print_err('Found unkown emotion %s' % edge)
            continue

        index[emotion] = edge['reaction_count'] or 0

    return index


def format_err(err):
    return [err] + ERROR_PADDING


def format(data):
    video_view_count = data.get('video_view_count')
    scraped = data['scraped'] or {}

    row = [
        '',
        data.get('url', ''),
        scraped.get('account_name', ''),
        scraped.get('timestamp', ''),
        scraped.get('time', ''),
        scraped.get('link', ''),
        scraped.get('aria_label', ''),
        scraped.get('text', ''),
        get_count(data['share_count']),
        get_count(data['comment_count']),
        get_count(data['reaction_count']),
        video_view_count if isinstance(video_view_count, int) else ''
    ]

    emotion_index = collect_top_reactions(data)

    for emotion_name in REACTION_KEYS.values():
        row.append(emotion_index.get(emotion_name, 0))

    return row


def facebook_post_stats_action(namespace):

    # Handling output
    output_file = open_output_file(namespace.output)

    enricher = casanova.enricher(
        namespace.file,
        output_file,
        add=REPORT_HEADERS,
        keep=namespace.select
    )

    http = create_pool()

    def fetch_facebook_page_stats(url):
        err, response = request(http, url, cookie='locale=en_US')

        if err:
            return 'http-error', None

        if response.status == 404:
            return 'not-found', None

        if response.status >= 400:
            return 'http-error', None

        html = response.data

        if CAPTCHA in html:
            die([
                'Rate limit reached!',
                'Last url: %s' % url
            ])

        if (
            CURRENT_AVAILABILITY_DISCLAIMER in html or
            AVAILABILITY_DISCLAIMER in html
        ):
            return 'unavailable', None

        if LOGIN_DISCLAIMER in html:
            return 'private-or-unavailable', None

        # TODO: integrate into ural
        bpost_id = url.rsplit('/', 1)[-1].encode()

        # Extracting metadata
        meta_extractor = re.compile(META_EXTRACTOR_TEMPLATE % bpost_id)

        match = meta_extractor.search(html)

        if match is None:
            return 'extraction-failed', None

        data = json5.loads(match.group(1).decode())
        data = nested_get(
            ['jsmods', 'pre_display_requires', 0, 3, 1, '__bbox', 'result', 'data', 'feedback'],
            data
        )

        if data is None:
            return 'extraction-failed', None

        # TODO: remove, this is here as a test
        # TODO: try to find a post where comments are disabled
        if get_count(data['seen_by_count']):
            print_err('Found seen_by_count: %i for %s' % (get_count(data['seen_by_count']), url))

        if 'political_figure_data' in data and data['political_figure_data']:
            print_err('Found political_figure_data:')
            print_err(data['political_figure_data'])

        if get_count(data['reaction_count']) != get_count(data['reactors']):
            print_err('Found different reactions/reactors for %s' % url)

        # Extracting data from hidden html
        hidden_html_extractor = re.compile(HTML_EXTRACTOR_TEMPLATE % bpost_id)
        match = hidden_html_extractor.search(html)

        if match is not None:
            hidden_html = match.group(1).decode()
            soup = BeautifulSoup(hidden_html, 'lxml')

            # Sometimes fetching a post behaves weirdly
            if soup.select_one('h5 a') is None:
                return 'extraction-failed', None

            data['scraped'] = {}

            timestamp_elem = soup.select_one('[data-utime]')
            timestamp = int(timestamp_elem.get('data-utime'))

            data['scraped']['account_name'] = soup.select_one('h5 a').get_text().strip()
            data['scraped']['timestamp'] = timestamp
            data['scraped']['time'] = datetime.fromtimestamp(timestamp).isoformat()

            # TODO: use a context manager
            try:
                data['scraped']['aria_label'] = timestamp_elem.parent.get('aria-label')
            except:
                pass

            try:
                data['scraped']['text'] = soup.select_one('[data-testid="post_message"]').get_text()
            except:
                pass

            # try:
            #     data['scraped']['link'] = soup.select_one('[data-lynx-uri]').get('href')
            # except:
            #     pass

        return None, data

    # Loading bar
    loading_bar = tqdm(
        desc='Fetching post stats',
        dynamic_ncols=True,
        unit=' posts',
        total=namespace.total
    )

    for row, post_url in enricher.cells(namespace.column, with_rows=True):
        loading_bar.update()

        if (
            not post_url or
            not is_facebook_post_url(post_url) or
            not is_facebook_url(post_url)
        ):
            enricher.writerow(row, format_err('not-facebook-post'))
            continue

        err, data = fetch_facebook_page_stats(post_url)

        if err:
            enricher.writerow(row, format_err(err))
        else:
            enricher.writerow(row, format(data))

        # Throttling
        sleep_with_entropy(FACEBOOK_WEB_DEFAULT_THROTTLE, 5.0)



# =============================================================================
# Minet CrowdTangle Links Summary
# =============================================================================
#
# Function related to link summary using CrowdTangle's APIs.
#
from urllib.parse import quote

from minet.crowdtangle.exceptions import (
    CrowdTangleMissingTokenError,
    CrowdTangleInvalidTokenError,
    CrowdTangleInvalidRequestError
)
from minet.utils import request_json, nested_get
from minet.crowdtangle.constants import (
    CROWDTANGLE_SUMMARY_DEFAULT_SORT_TYPE,
    CROWDTANGLE_SUMMARY_SORT_TYPES,
    CROWDTANGLE_OUTPUT_FORMATS
)
from minet.crowdtangle.formatters import (
    format_post,
    format_summary
)

URL_TEMPLATE = (
    'https://api.crowdtangle.com/links'
    '?token=%(token)s'
    '&count=%(count)s'
    '&startDate=%(start_date)s'
    '&includeSummary=true'
    '&link=%(link)s'
    '&sortBy=%(sort_by)s'
)


def url_forge(link, token, start_date, sort_by, include_posts=False):
    return URL_TEMPLATE % {
        'token': token,
        'count': 1 if not include_posts else 100,
        'start_date': start_date,
        'link': quote(link, safe=''),
        'sort_by': sort_by
    }


def crowdtangle_summary(http, link, token=None, start_date=None, with_top_posts=False,
                        sort_by=CROWDTANGLE_SUMMARY_DEFAULT_SORT_TYPE, format='csv_dict_row'):

    if token is None:
        raise CrowdTangleMissingTokenError

    if format not in CROWDTANGLE_OUTPUT_FORMATS:
        raise TypeError('minet.crowdtangle.summary: unkown `format`.')

    if not isinstance(start_date, str):
        raise TypeError('minet.crowdtangle.summary: expecting a `start_date` kwarg.')

    if sort_by not in CROWDTANGLE_SUMMARY_SORT_TYPES:
        raise TypeError('minet.crowdtangle.summary: unknown `sort_by`.')

    # Fetching
    api_url = url_forge(
        link,
        token,
        start_date,
        sort_by,
        with_top_posts
    )

    err, response, data = request_json(http, api_url)

    if err is not None:
        raise err

    if response.status == 401:
        raise CrowdTangleInvalidTokenError

    if response.status >= 400:
        raise CrowdTangleInvalidRequestError(api_url)

    stats = nested_get(['result', 'summary', 'facebook'], data)
    posts = nested_get(['result', 'posts'], data) if with_top_posts else None

    if stats is not None:
        if format == 'csv_dict_row':
            stats = format_summary(stats, as_dict=True)
        elif format == 'csv_row':
            stats = format_summary(stats)

    if not with_top_posts:
        return stats

    else:
        if posts is not None:
            if format == 'csv_dict_row':
                posts = [format_post(post, as_dict=True) for post in posts]
            elif format == 'csv_row':
                posts = [format_post(post) for post in posts]

        return stats, posts



# =============================================================================
# Minet Mediacloud Constants
# =============================================================================
#
# Bunch of mediacloud-related constants.
#
from urllib3 import Timeout


MEDIACLOUD_API_BASE_URL = 'https://api.mediacloud.org/api/v2'
MEDIACLOUD_DEFAULT_TIMEOUT = Timeout(connect=30, read=60 * 5)
MEDIACLOUD_DEFAULT_BATCH = 250

MEDIACLOUD_TOPIC_STORIES_CSV_HEADERS = [
    'guid',
    'stories_id',
    'title',
    'url',
    'language',
    'media_id',
    'media_name',
    'collect_date',
    'publish_date',
    'date_is_reliable',
    'facebook_share_count',
    'full_text_rss',
    'inlink_count',
    'outlink_count',
    'media_inlink_count',
    'post_count',
    'snapshots_id',
    'timespans_id',
    'next_link_id'
]

MEDIACLOUD_STORIES_CSV_HEADER = [
    'guid',
    'stories_id',
    'processed_stories_id',
    'title',
    'url',
    'language',
    'collect_date',
    'publish_date',
    'media_id',
    'media_name',
    'media_url',
    'tags',
    'tag_sets',
    'tags_ids',
    'tag_sets_ids'
]



# =============================================================================
# Minet Extract Content CLI Action
# =============================================================================
#
# Logic of the extract action.
#
import casanova
import gzip
import codecs
import warnings
from multiprocessing import Pool
from tqdm import tqdm
from dragnet import extract_content

from minet.encodings import is_supported_encoding
from minet.cli.utils import (
    open_output_file,
    create_report_iterator
)
from minet.cli.reporters import report_error

from minet.exceptions import UnknownEncodingError

OUTPUT_ADDITIONAL_HEADERS = ['extract_error', 'extracted_text']


def worker(payload):
    row, _, path, encoding, content, _ = payload

    if not is_supported_encoding(encoding):
        return UnknownEncodingError('Unknown encoding: "%s"' % encoding), row, None

    # Reading file
    if content is None:
        try:
            if path.endswith('.gz'):
                with open(path, 'rb') as f:
                    raw_html_bytes = gzip.decompress(f.read())

                raw_html = raw_html_bytes.decode(encoding, errors='replace')
            else:
                with codecs.open(path, 'r', encoding=encoding, errors='replace') as f:
                    raw_html = f.read()
        except UnicodeDecodeError as e:
            return e, row, None
    else:
        raw_html = content

    # Attempting extraction
    try:
        with warnings.catch_warnings():
            warnings.simplefilter('ignore')
            content = extract_content(raw_html)
    except BaseException as e:
        return e, row, None

    return None, row, content


def extract_action(namespace):
    output_file = open_output_file(namespace.output)

    enricher = casanova.enricher(
        namespace.report,
        output_file,
        keep=namespace.select,
        add=OUTPUT_ADDITIONAL_HEADERS
    )

    loading_bar = tqdm(
        desc='Extracting content',
        total=namespace.total,
        dynamic_ncols=True,
        unit=' docs'
    )

    files = create_report_iterator(namespace, enricher, loading_bar=loading_bar)

    with Pool(namespace.processes) as pool:
        for error, row, content in pool.imap_unordered(worker, files):
            loading_bar.update()

            if error is not None:
                enricher.writerow(row, [report_error(error), ''])
                continue

            enricher.writerow(row, ['', content])

    output_file.close()



# =============================================================================
# Minet Facebook Url Likes CLI Action
# =============================================================================
#
# Action reading an input CSV file line by line and retrieving approximate
# likes count by scraping Facebook's like button plugin.
#
import casanova
from tqdm import tqdm

from minet.cli.utils import open_output_file, die

REPORT_HEADERS = ['approx_likes', 'approx_likes_int']


def facebook_url_likes_action(namespace):
    output_file = open_output_file(namespace.output)

    enricher = casanova.enricher(
        namespace.file,
        output_file,
        keep=namespace.select,
        add=REPORT_HEADERS
    )

    if namespace.column not in enricher.pos:
        die([
            'Could not find the "%s" column containing the urls in the given CSV file.' % namespace.column
        ])

    loading_bar = tqdm(
        desc='Retrieving likes',
        dynamic_ncols=True,
        unit=' urls',
    )

    for row, url in enricher.cells(namespace.column, with_rows=True):

        loading_bar.update()

        url_data = url.strip()

        enricher.writerow(row)



# =============================================================================
# Minet CrowdTangle Posts
# =============================================================================
#
# Function related to posts fetching.
#
from minet.crowdtangle.utils import make_paginated_iterator
from minet.crowdtangle.formatters import format_post

URL_TEMPLATE = 'https://api.crowdtangle.com/posts?count=100&sortBy=%(sort_by)s&token=%(token)s'


def url_forge(token=None, sort_by=None, language=None, start_date=None,
              end_date=None, list_ids=None, **kwargs):

    base_url = URL_TEMPLATE % {
        'sort_by': sort_by,
        'token': token
    }

    if language:
        base_url += '&language=%s' % language

    if start_date:
        base_url += '&startDate=%s' % start_date

    if end_date:
        base_url += '&endDate=%s' % end_date

    if list_ids:
        base_url += '&listIds=%s' % ','.join(list_ids)

    return base_url


crowdtangle_posts = make_paginated_iterator(
    url_forge,
    item_key='posts',
    formatter=format_post
)



# =============================================================================
# Minet Mediacloud Search
# =============================================================================
#
# Function related to stories searching.
#
from urllib.parse import quote_plus

from minet.utils import request_json
from minet.mediacloud.constants import (
    MEDIACLOUD_API_BASE_URL,
    MEDIACLOUD_DEFAULT_BATCH
)
from minet.mediacloud.exceptions import (
    MediacloudServerError
)
from minet.mediacloud.formatters import format_story
from minet.mediacloud.utils import get_last_processed_stories_id


def query_additions(query, collections=None):
    if collections is not None:
        query += ' AND ('
        query += ' OR '.join('tags_id_media:%s' % str(collection) for collection in collections)
        query += ')'

    return query


def url_forge(token, query, collections=None, count=False,
              last_processed_stories_id=None):

    url = '%s/stories_public/%s?key=%s' % (
        MEDIACLOUD_API_BASE_URL,
        'count' if count else 'list',
        token
    )

    query = query_additions(query, collections=collections)

    url += '&q=%s' % quote_plus(query)

    if not count:
        url += '&rows=%i' % MEDIACLOUD_DEFAULT_BATCH

        if last_processed_stories_id is not None:
            url += '&last_processed_stories_id=%i' % last_processed_stories_id

    return url


def mediacloud_search(http, token, query, count=False, collections=None, format='csv_dict_row'):

    def generator():
        last_processed_stories_id = None

        while True:
            url = url_forge(
                token,
                query,
                collections=collections,
                count=count,
                last_processed_stories_id=last_processed_stories_id
            )

            err, response, data = request_json(http, url)

            if err:
                raise err

            if response.status >= 500:
                raise MediacloudServerError(server_error=data.get('error'))

            if count:
                yield data['count']
                return

            for story in data:
                if format == 'csv_dict_row':
                    yield format_story(story, as_dict=True)
                elif format == 'csv_row':
                    yield format_story(story)
                else:
                    yield story

            last_processed_stories_id = get_last_processed_stories_id(data)

            if last_processed_stories_id is None:
                return

    if count:
        return next(generator())

    return generator()



# =============================================================================
# Minet Url Extract CLI Action
# =============================================================================
#
# Logic of the `url-extract` action.
#
import casanova
from tqdm import tqdm
from ural import urls_from_text, urls_from_html
from urllib.parse import urljoin

from minet.cli.utils import open_output_file

REPORT_HEADERS = [
    'url'
]

EXTRACTORS = {
    'html': urls_from_html,
    'text': urls_from_text
}


def url_extract_action(namespace):
    output_file = open_output_file(namespace.output)

    enricher = casanova.enricher(
        namespace.file,
        output_file,
        add=REPORT_HEADERS,
        keep=namespace.select.split(',') if namespace.select else None
    )

    extract = EXTRACTORS[getattr(namespace, 'from')]

    loading_bar = tqdm(
        desc='Extracting',
        dynamic_ncols=True,
        unit=' rows',
        total=namespace.total
    )

    for row, content in enricher.cells(namespace.column, with_rows=True):
        loading_bar.update()

        content = content.strip()

        if not content:
            continue

        for url in extract(content):
            if namespace.base_url is not None:
                url = urljoin(namespace.base_url, url)

            enricher.writerow(row, [url])

    output_file.close()



# =============================================================================
# Minet Fetch CLI Action
# =============================================================================
#
# Action reading an input CSV file line by line and fetching the urls found
# in the given column. This is done in a respectful multithreaded fashion to
# optimize both running time & memory.
#
import os
import sys
import gzip
import casanova
from io import StringIO
from os.path import join, dirname, isfile
from collections import Counter
from tqdm import tqdm
from uuid import uuid4
from ural import is_url

from minet.fetch import multithreaded_fetch
from minet.utils import (
    grab_cookies,
    parse_http_header,
    PseudoFStringFormatter
)
from minet.cli.reporters import report_error
from minet.cli.utils import (
    open_output_file,
    die,
    LazyLineDict
)

OUTPUT_ADDITIONAL_HEADERS = [
    'resolved',
    'status',
    'error',
    'filename',
    'encoding'
]

CUSTOM_FORMATTER = PseudoFStringFormatter()


def fetch_action(namespace):

    # Are we resuming
    resuming = namespace.resume

    if resuming and not namespace.output:
        die([
            'Cannot --resume without specifying -o/--output.'
        ])

    # Do we need to fetch only a single url?
    single_url = namespace.file is sys.stdin and is_url(namespace.column)

    if single_url:
        namespace.file = StringIO('url\n%s' % namespace.column)
        namespace.column = 'url'

        # If we are hitting a single url we enable contents_in_report
        if namespace.contents_in_report is None:
            namespace.contents_in_report = True

    # HTTP method
    http_method = namespace.method

    # Cookie grabber
    get_cookie = None
    if namespace.grab_cookies:
        get_cookie = grab_cookies(namespace.grab_cookies)

    # Global headers
    global_headers = None
    if namespace.headers:
        global_headers = {}

        for header in namespace.headers:
            k, v = parse_http_header(header)
            global_headers = v

    flag = 'w'
    if namespace.output is not None and resuming and isfile(namespace.output):
        flag = 'r+'

    output_file = open_output_file(namespace.output, flag=flag)

    # Resume listener
    listener = None
    resuming_reader_loading = None
    skipped = 0

    if resuming:
        resuming_reader_loading = tqdm(
            desc='Resuming',
            dynamic_ncols=True,
            unit=' lines'
        )

        def listener(event, row):
            nonlocal skipped

            if event == 'resume.output':
                resuming_reader_loading.update()

            if event == 'resume.input':
                skipped += 1
                loading_bar.set_postfix(skipped=skipped)
                loading_bar.update()

    # Enricher
    enricher = casanova.threadsafe_enricher(
        namespace.file,
        output_file,
        resumable=resuming,
        auto_resume=False,
        add=OUTPUT_ADDITIONAL_HEADERS + (['raw_contents'] if namespace.contents_in_report else []),
        keep=namespace.select,
        listener=listener
    )

    if namespace.column not in enricher.pos:
        die([
            'Could not find the "%s" column containing the urls in the given CSV file.' % namespace.column
        ])

    url_pos = enricher.pos[namespace.column]

    filename_pos = None

    if namespace.filename is not None:
        if namespace.filename not in enricher.pos:
            die([
                'Could not find the "%s" column containing the filenames in the given CSV file.' % namespace.filename
            ])

        filename_pos = enricher.pos[namespace.filename]

    indexed_input_headers = {h: i for i, h in enumerate(enricher.fieldnames)}

    if resuming:
        enricher.resume()
        resuming_reader_loading.close()

    # Loading bar
    total = namespace.total

    loading_bar = tqdm(
        desc='Fetching pages',
        total=total,
        dynamic_ncols=True,
        unit=' urls'
    )

    def url_key(item):
        url = item[1][url_pos].strip()

        if not url:
            return

        # Url templating
        if namespace.url_template:
            return namespace.url_template.format(value=url)

        return url

    def request_args(url, item):
        cookie = None

        # Cookie
        if get_cookie:
            cookie = get_cookie(url)

        # Headers
        headers = None

        if global_headers:
            headers = global_headers

        return {
            'method': http_method,
            'cookie': cookie,
            'headers': headers
        }

    def write_output(index, row, resolved=None, status=None, error=None,
                     filename=None, encoding=None, data=None):

        addendum = [
            resolved or '',
            status or '',
            error or '',
            filename or '',
            encoding or ''
        ]

        if namespace.contents_in_report:
            addendum.append(data or '')

        enricher.writerow(index, row, addendum)

    errors = 0
    status_codes = Counter()

    multithreaded_iterator = multithreaded_fetch(
        enricher,
        key=url_key,
        request_args=request_args,
        threads=namespace.threads,
        throttle=namespace.throttle,
        domain_parallelism=namespace.domain_parallelism
    )

    for result in multithreaded_iterator:
        index, row = result.item

        if not result.url:

            write_output(
                index,
                row
            )

            loading_bar.update()
            continue

        response = result.response
        data = response.data if response is not None else None

        content_write_flag = 'wb'

        # Updating stats
        if result.error is not None:
            errors += 1
        else:
            if response.status >= 400:
                status_codes[response.status] += 1

        postfix = {'errors': errors}

        for code, count in status_codes.most_common(1):
            postfix[str(code)] = count

        loading_bar.set_postfix(**postfix)
        loading_bar.update()

        # No error
        if result.error is None:

            filename = None

            # Building filename
            if data:
                if filename_pos is not None or namespace.filename_template:
                    if namespace.filename_template:
                        filename = CUSTOM_FORMATTER.format(
                            namespace.filename_template,
                            value=row[filename_pos] if filename_pos is not None else None,
                            ext=result.meta['ext'],
                            line=LazyLineDict(indexed_input_headers, row)
                        )
                    else:
                        filename = row[filename_pos] + result.meta['ext']
                else:
                    # NOTE: it would be nice to have an id that can be sorted by time
                    filename = str(uuid4()) + result.meta['ext']

            # Standardize encoding?
            encoding = result.meta['encoding']

            if data and namespace.standardize_encoding or namespace.contents_in_report:
                if encoding is None or encoding != 'utf-8' or namespace.contents_in_report:
                    data = data.decode(encoding if encoding is not None else 'utf-8', errors='replace')
                    encoding = 'utf-8'
                    content_write_flag = 'w'

            # Writing file on disk
            if data and not namespace.contents_in_report:

                if namespace.compress:
                    filename += '.gz'

                resource_path = join(namespace.output_dir, filename)
                resource_dir = dirname(resource_path)

                os.makedirs(resource_dir, exist_ok=True)

                with open(resource_path, content_write_flag) as f:

                    # TODO: what if standardize_encoding + compress?
                    f.write(gzip.compress(data) if namespace.compress else data)

            # Reporting in output
            resolved_url = response.geturl()

            write_output(
                index,
                row,
                resolved=resolved_url if resolved_url != result.url else None,
                status=response.status,
                filename=filename,
                encoding=encoding,
                data=data
            )

        # Handling potential errors
        else:
            error_code = report_error(result.error)

            write_output(
                index,
                row,
                error=error_code
            )

    # Closing files
    output_file.close()



# =============================================================================
# Minet Facebook Post Id From Url
# =============================================================================
#
# Helper used to retrieved a full facebook post id from the given post url.
#
import json
from bs4 import BeautifulSoup
from urllib.parse import urlsplit, parse_qsl, urljoin
from ural.facebook import (
    convert_facebook_url_to_mobile,
    parse_facebook_url,
    FacebookPost
)

from minet.utils import rate_limited_from_state, request_text
from minet.facebook.constants import (
    FACEBOOK_URL,
    FACEBOOK_MOBILE_RATE_LIMITER_STATE,
    FACEBOOK_DEFAULT_POOL
)


@rate_limited_from_state(FACEBOOK_MOBILE_RATE_LIMITER_STATE)
def post_id_from_url(post_url):
    parsed = parse_facebook_url(post_url)

    if not isinstance(parsed, FacebookPost):
        return

    if parsed.full_id is not None:
        return parsed.full_id

    post_mobile_url = convert_facebook_url_to_mobile(post_url)

    err, response, html = request_text(FACEBOOK_DEFAULT_POOL, post_mobile_url)

    if err:
        raise err

    soup = BeautifulSoup(html, 'lxml')

    root_element = soup.select_one('#m_story_permalink_view [data-ft]')

    if root_element is None:

        # Is this a photo post?
        next_link = soup.select_one('[href^="/photo.php"]')

        if next_link is None:
            return

        href = next_link.get('href')

        if not href:
            return

        link = urljoin(FACEBOOK_URL, href)
        query = urlsplit(link).query

        if not query:
            return

        query = dict(parse_qsl(query))

        return '%s_%s' % (query['id'], query['fbid'])

    data = root_element.get('data-ft')

    if data is None:
        return

    try:
        data = json.loads(data)
    except json.JSONDecodeError:
        return

    content_owner_id_new = data.get('content_owner_id_new') or data.get('page_id')
    mf_story_key = data.get('mf_story_key')

    if content_owner_id_new is None or mf_story_key is None:
        return

    return '%s_%s' % (content_owner_id_new, mf_story_key)



from minet.__version__ import __version__

from minet.crawl import (
    crawl,
    Crawler,
    CrawlJob,
    Spider,
    BeautifulSoupSpider,
    DefinitionSpider
)
from minet.fetch import multithreaded_fetch, multithreaded_resolve
from minet.scrape import scrape, Scraper
from minet.utils import (
    RateLimiter,
    RateLimiterState,
    RateLimitedIterator,
    RetryableIterator,
    rate_limited,
    rate_limited_method
)



from setuptools import setup, find_packages

with open('./README.md', 'r') as f:
    long_description = f.read()

# Version
# Info: https://packaging.python.org/guides/single-sourcing-package-version/
# Example: https://github.com/pypa/warehouse/blob/64ca42e42d5613c8339b3ec5e1cb7765c6b23083/warehouse/__about__.py
meta_package = {}
with open('./minet/__version__.py') as f:
  exec(f.read(), meta_package)


setup(name='minet',
      version=meta_package['__version__'],
      description='A webmining CLI tool & library for python.',
      long_description=long_description,
      long_description_content_type='text/markdown',
      url='http://github.com/medialab/minet',
      license='MIT',
      author='Jules Farjas, Guillaume Plique, Pauline Breteau',
      keywords='webmining',
      python_requires='>=3',
      packages=find_packages(exclude=['ftest', 'scripts', 'test']),
      install_requires=[
        'beautifulsoup4>=4.7.1',
        'browser-cookie3==0.7.6',
        'casanova==0.9.0',
        'cchardet==2.1.4',
        'cython>=0.29.4',
        'dateparser>=0.7.1',
        'json5>=0.8.5',
        'keyring<19.3',
        'lxml>=4.3.0',
        'ndjson>=0.3.1',
        'numpy>=1.16.1',
        'persist-queue>=0.4.2',
        'pytz>=2019.3',
        'pyyaml',
        'quenouille>=0.6.2',
        'tqdm>=4.31.1',
        'twitter>=1.18.0',
        'ural>=0.25.0',
        'urllib3[secure]>=1.25.3'
      ],
      entry_points={
        'console_scripts': ['minet=minet.cli.__main__:main']
      },
      zip_safe=True)



# =============================================================================
# Minet CLI Reporters
# =============================================================================
#
# Various reporters whose goal is to convert errors etc. into human-actionable
# labels in CSV format, for instance.
#
from urllib3.exceptions import (
    ConnectTimeoutError,
    MaxRetryError,
    ReadTimeoutError,
    ResponseError,
    SSLError,
    NewConnectionError,
    ProtocolError
)

from minet.exceptions import (
    UnknownEncodingError,
    InvalidURLError
)


def max_retry_error_reporter(error):
    if isinstance(error, ConnectTimeoutError):
        return 'connect-timeout'

    if isinstance(error, ReadTimeoutError):
        return 'read-timeout'

    if isinstance(error.reason, ResponseError) and 'redirect' in repr(error.reason):
        return 'too-many-redirects'

    return 'max-retries-exceeded'


def new_connection_error_reporter(error):
    msg = repr(error)

    if 'Name or service not known' in msg:
        return 'unknown-dns'

    return msg


def protocol_error_reporter(error):
    msg = repr(error)

    if 'Connection aborted' in msg:
        return 'connection-aborted'

    return msg


ERROR_REPORTERS = {
    UnicodeDecodeError: 'wrong-encoding',
    UnknownEncodingError: 'unknown-encoding',
    MaxRetryError: max_retry_error_reporter,
    InvalidURLError: 'invalid-url',
    SSLError: 'ssl',
    NewConnectionError: new_connection_error_reporter,
    ProtocolError: protocol_error_reporter,
    ConnectTimeoutError: 'connect-timeout',
    ReadTimeoutError: 'read-timeout'
}


def report_error(error):
    reporter = ERROR_REPORTERS.get(type(error), repr)

    return reporter(error) if callable(reporter) else reporter



# =============================================================================
# Minet CrowdTangle Leaderboard CLI Action
# =============================================================================
#
# Logic of the `ct leaderboard` action.
#
from minet.crowdtangle.constants import (
    CROWDTANGLE_LEADERBOARD_CSV_HEADERS,
    CROWDTANGLE_LEADERBOARD_CSV_HEADERS_WITH_BREAKDOWN
)
from minet.cli.crowdtangle.utils import make_paginated_action


def select_csv_headers(namespace):
    if namespace.breakdown:
        return CROWDTANGLE_LEADERBOARD_CSV_HEADERS_WITH_BREAKDOWN

    return CROWDTANGLE_LEADERBOARD_CSV_HEADERS


crowdtangle_leaderboard_action = make_paginated_action(
    method_name='leaderboard',
    item_name='accounts',
    csv_headers=select_csv_headers
)



# =============================================================================
# Minet Twitter CLI Action
# =============================================================================
#
# Logic of the `tw` action.
#
import sys

from minet.cli.utils import open_output_file, die


def twitter_action(namespace):

    # Credentials are required to be able to access the API
    if not namespace.api_key or \
       not namespace.api_secret_key or \
       not namespace.access_token or \
       not namespace.access_token_secret:
        die([
            'Full credentials are required to access Twitter API.',
            'You can provide them using various CLI arguments:',
            '    --api-key',
            '    --api-secret-key',
            '    --access-token',
            '    --access-token-secret'
        ])

    output_file = open_output_file(
        namespace.output,
        flag='a+' if getattr(namespace, 'resume', False) else 'w'
    )

    if namespace.tw_action == 'friends':
        from minet.cli.twitter.friends import twitter_friends_action

        twitter_friends_action(namespace, output_file)

    elif namespace.tw_action == 'followers':
        from minet.cli.twitter.followers import twitter_followers_action

        twitter_followers_action(namespace, output_file)

    elif namespace.tw_action == 'users':
        from minet.cli.twitter.users import twitter_users_action

        twitter_users_action(namespace, output_file)

    # Cleanup
    if namespace.output is not None:
        output_file.close()
